# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# -*- coding: utf-8 -*-
import os
import tempfile
import dataiku
from datasets import Dataset
import pandas as pd
import torch
from transformers import AutoModelForCausalLM, AutoTokenizer, BitsAndBytesConfig, TrainingArguments
from trl import DPOTrainer
from peft import LoraConfig

MODEL_NAME = "mistralai/Mistral-7B-Instruct-v0.2"
MODEL_REVISION = "c72e5d1908b1e2929ec8fc4c8820e9706af1f80f"

train_dataset = Dataset.from_pandas(
    dataiku.Dataset("po_train").get_dataframe()
)
validation_dataset = Dataset.from_pandas(
    dataiku.Dataset("po_validation").get_dataframe()
)

folder = dataiku.Folder("HUY9JaOD")

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
bnb_config = BitsAndBytesConfig(
    load_in_4bit=True,
    bnb_4bit_quant_type="nf4",
    bnb_4bit_compute_dtype=torch.float16,
)

model = AutoModelForCausalLM.from_pretrained(
    MODEL_NAME,
    revision=MODEL_REVISION,
    device_map="auto",
    quantization_config=bnb_config,
)
model.config.use_cache = False # Because the model will change as it is fine-tuned

tokenizer = AutoTokenizer.from_pretrained(
    MODEL_NAME,
    revision=MODEL_REVISION
)
tokenizer.pad_token = tokenizer.eos_token

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
def return_prompt_and_responses(samples):
    """
    Transform a batch of examples in a format suitable for DPO.
    """
    return {
        "prompt": [
            f"[INST] Extract key details from: {question} [/INST] "
            for question in samples["question"]
        ],
        "chosen": samples["chosen"],
        "rejected": samples["rejected"]
    }

def transform(ds):
    """
    Prepare the datasets in a format suitable for DPO.
    """
    return ds.map(
        return_prompt_and_responses,
        batched=True,
        remove_columns=ds.column_names
    )

train_dataset = transform(train_dataset)
validation_dataset = transform(validation_dataset)

with tempfile.TemporaryDirectory() as temp_dir:

    # Define the training parameters
    training_args = TrainingArguments(
        per_device_train_batch_size=4,
        num_train_epochs=10,
        optim="paged_adamw_8bit",
        gradient_checkpointing=True,
        output_dir=temp_dir,
    )
    
    peft_config = LoraConfig(
        r=16,
        lora_alpha=32,
        lora_dropout=0.05,
        task_type="CAUSAL_LM",
    )    

    dpo_trainer = DPOTrainer(
        model,
        None, # The reference model is the base model (without LoRA adaptation)
        peft_config=peft_config,
        args=training_args,
        beta=0.005,
        loss_type="ipo",
        train_dataset=train_dataset,
        eval_dataset=validation_dataset,
        tokenizer=tokenizer,
    )
    
    # Fine-tune the model
    dpo_trainer.train()
    
    # Save the fine-tuned model in the managed folder
    dpo_trainer.save_model(output_dir=temp_dir)
    for root, dirs, files in os.walk(temp_dir):
        for file in files:
            source_path = os.path.join(root, file)
            target_path = os.path.relpath(source_path, temp_dir)
            folder.upload_file(target_path, source_path)