# -------------------------------------------------------------------------------- NOTEBOOK-CELL: MARKDOWN
# # Agent Evaluation Method #1: LLM-as-a-judge with `langchain`
# Required inputs:
# - request
# - reply
# - intermediate steps

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
import dataiku
from dataiku.langchain.dku_llm import DKUChatLLM
from langchain.evaluation import load_evaluator
from langchain_core.agents import AgentAction
import json

LLM_ID = dataiku.get_custom_variables()["LLM_id"]
llm = DKUChatLLM(
    llm_id=LLM_ID,
    temperature=0
)
evaluator = load_evaluator("trajectory", llm=llm)

df = dataiku.Dataset("requests_processed_stacked").get_dataframe()

def deserialize_intermediate_steps(s):
    """
    Deserialize the string representing the intermediate steps of an agent trajectory.
    """
    return [
        (
            AgentAction(tool=action[0], tool_input=action[1], log=action[2]),
            action[3]
        )
        for action in json.loads(s)
    ]

for i in df.index:    
    result = evaluator.evaluate_agent_trajectory(
        prediction=df.at[i, "draft_reply"],
        input=df.at[i, "request"],
        agent_trajectory=deserialize_intermediate_steps(df.at[i, "intermediate_steps"]),
    )
    df.at[i, "langchain/score"] = result["score"]
    df.at[i, "langchain/justification"] = result["reasoning"]

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: MARKDOWN
# # Agent Evaluation Method #2: LLM-as-a-judge with `mlflow`
# Required inputs:
# - request
# - reply
# - ground truth reply

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
import mlflow
from mlflow.metrics import genai
import os

auth_info = dataiku.api_client().get_auth_info(with_secrets=True)
for secret in auth_info["secrets"]:
    if secret["key"] == "openai_key":
        os.environ["OPENAI_API_KEY"] = secret["value"]
        break

with mlflow.start_run() as run:
    results = mlflow.evaluate(
        data=df,
        targets="reference_answer",
        predictions="draft_reply",
        extra_metrics=[
            genai.answer_correctness(),
        ],
        evaluators="default",
        evaluator_config={'col_mapping': {'inputs': 'request'}}
    )

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
df = results.tables['eval_results_table']
df = df[['customer_id', 'request', 'reference_answer', 'reference_trajectory', 'agent', 'draft_reply', 'actions', 'intermediate_steps', 'langchain/score', 'langchain/justification', 'answer_correctness/v1/score', 'answer_correctness/v1/justification']]
df = df.rename(
    {
        "answer_correctness/v1/justification": "mlflow/justification",
        "answer_correctness/v1/score": "mlflow/score"
    },
    axis=1
)
dataiku.Dataset("agent_answers_evaluated").write_with_schema(df)