# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
import os
import io
import json
from datetime import datetime

import dataiku
from dataikuapi.dss.modelevaluationstore import DSSModelEvaluationStore
import pandas as pd
import mlflow
from mlflow.metrics import genai, make_metric
from mlflow.metrics.base import MetricValue, standard_aggregations

import evaluate

auth_info = dataiku.api_client().get_auth_info(with_secrets=True)
for secret in auth_info["secrets"]:
    if secret["key"] == "openai_key":
        os.environ["OPENAI_API_KEY"] = secret["value"]
        break

EVALUATION_STORE_ID = "jNuKpcAc"
MLFLOW_FOLDER_ID = "quL4tFAd"

EXPERIMENT_NAME = "question_answering_code"
project = dataiku.api_client().get_default_project()
mlflow_handle = project.setup_mlflow(project.get_managed_folder(MLFLOW_FOLDER_ID))
mlflow.set_experiment(EXPERIMENT_NAME)

df = dataiku.Dataset("answers2").get_dataframe()

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
bertscore = evaluate.load("bertscore", keep_in_memory=True)

def bert_score_fn(predictions, targets, metrics):
    scores = bertscore.compute(
        predictions=predictions,
        references=targets,
        model_type="distilbert-base-uncased"
    )["f1"]
    return MetricValue(
        scores=scores,
        aggregate_results=standard_aggregations(scores),
    )

bert_score = make_metric(
    eval_fn=bert_score_fn, greater_is_better=True, name="BERT score"
)

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
with mlflow.start_run() as run:
    results = mlflow.evaluate(
        data=df,
        targets="reference_answer",
        predictions="answer",
        extra_metrics=[
            bert_score,
            genai.answer_correctness(),
            genai.answer_relevance(),
            genai.relevance(),
            genai.faithfulness()
        ],
        evaluators="default",
        evaluator_config={'col_mapping': {'inputs': 'question'}}
    )

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
evaluation_df = None

folder = dataiku.Folder(MLFLOW_FOLDER_ID)

for path in folder.list_paths_in_partition():
    if "eval_results_table.json" in path:
        with folder.get_download_stream(path) as stream:
            buf = io.BytesIO(stream.read())
            d = json.loads(buf.read())
            if evaluation_df is None:
                evaluation_df = pd.DataFrame.from_dict(d["data"])
                evaluation_df.columns = d["columns"]
            else:
                new_df = pd.DataFrame.from_dict(d["data"])
                new_df.columns = d["columns"]
                evaluation_df = pd.concat((evaluation_df, new_df), axis=0, ignore_index=True)

dataiku.Dataset("answers_evaluated2").write_with_schema(evaluation_df)

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
scores = []

for metric in results.metrics:
    if "/mean" in metric:
        name = metric.split("/mean")[0]
    scores.append(
        DSSModelEvaluationStore.MetricDefinition(
            code=name,
            value=results.metrics[metric],
            name=metric,
            description=metric
        )
    )

eval_timestamp = datetime.now().isoformat()
label = DSSModelEvaluationStore.LabelDefinition("evaluation:date", eval_timestamp)
mes = dataiku.api_client().get_default_project().get_model_evaluation_store(EVALUATION_STORE_ID)
mes.add_custom_model_evaluation(scores, labels=[label])