# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
import dataiku
from dataiku.langchain.dku_llm import DKUChatLLM
import re
from langchain.chains.constitutional_ai.models import ConstitutionalPrinciple
from langchain.chains.constitutional_ai.principles import PRINCIPLES
from langchain.chains.constitutional_ai.prompts import CRITIQUE_PROMPT, REVISION_PROMPT
from langchain.chains.llm import LLMChain

df = dataiku.Dataset("movie_quotes").get_dataframe()        

LLM_ID = dataiku.get_custom_variables()["LLM_id"]
llm = DKUChatLLM(llm_id=LLM_ID, temperature=0)

critique_chain = LLMChain(llm=llm, prompt=CRITIQUE_PROMPT)
revision_chain = LLMChain(llm=llm, prompt=REVISION_PROMPT)

input_prompt = "Provide a movie quote I could send to a friend"

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
constitutional_principles = [
    x[1]
    for x in PRINCIPLES.items()
    if x[1].name in ["harmful1", "offensive", "illegal"]
]

def parse_critique(output_string):
    """
    Detect whether a revision request is included in the text generated by the LLM.
    """
    if "Revision request:" not in output_string:
        return output_string
    output_string = output_string.split("Revision request:")[0]
    if "\n\n" in output_string:
        output_string = output_string.split("\n\n")[0]
    return output_string

def critique_and_revise(response):
    """
    Given an initial response, go through each principle, criticize it and update it if needed.
    """
    critiques_and_revisions = []
    for constitutional_principle in constitutional_principles:
        # Criticize the response
        raw_critique = critique_chain.run(
            input_prompt=input_prompt,
            output_from_model=response,
            critique_request=constitutional_principle.critique_request
        )
        critique = parse_critique(output_string=raw_critique).strip()

        # If the critique contains "No critique needed", the response remains unchanged.
        for s in critique.lower().split("."):
            if re.search(r"\bno\b.*\bcritique\b.*\bneeded\b", s):
                critiques_and_revisions.append((critique, ""))
                break
        else:
            # Revise the response
            revision = revision_chain.run(
                input_prompt=input_prompt,
                output_from_model=response,
                critique_request=constitutional_principle.critique_request,
                critique=critique,
                revision_request=constitutional_principle.revision_request
            ).strip()
            if not re.search(r"\bno\b.*\brevisions?\b.*\bneeded\b", revision.lower()):
                response = revision
            critiques_and_revisions.append((critique, revision))

    explanation = ""

    for i in range(len(critiques_and_revisions)):
        explanation += f"\n\nPrinciple: '{constitutional_principles[i].name}'"
        x, y = critiques_and_revisions[i]
        explanation += f"\nCritique: {x}\nRevision: {y}"

    return response, explanation[2:]

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
for i in df.index:
    response, explanation = critique_and_revise(df.loc[i].text)
    df.at[i, "revised"] = response
    df.at[i, "explanation"] = explanation

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
dataiku.Dataset("moderation_constitutional-AI").write_with_schema(df)