# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
import dataiku
from dataiku.langchain.dku_llm import DKUChatLLM
import json
import mlflow
from spacy import displacy

from langchain.agents import AgentExecutor, create_structured_chat_agent
from langchain_core.messages import AIMessage, HumanMessage, SystemMessage
from langchain_core.prompts import ChatPromptTemplate, PromptTemplate, HumanMessagePromptTemplate
from langchain_core.prompts import MessagesPlaceholder
from langchain.tools import StructuredTool
from langchain.callbacks import MlflowCallbackHandler

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
df = dataiku.Dataset("requests").get_dataframe()

LLM_ID = dataiku.get_custom_variables()["LLM_id"]

project = dataiku.api_client().get_default_project()
mlflow_handle = project.setup_mlflow(project.get_managed_folder("qScFHunU"))
EXPERIMENT_NAME = "agent_chat_model"
mlflow.set_experiment(EXPERIMENT_NAME)

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
from fictitious_tools import (
    get_customer_id,
    get_details,
    reset_password,
    cancel_appointment,
    schedule_local_intervention,
    schedule_distant_intervention,
    run_diagnostics,
    sign_up_to_option,
    cancel_option
)

functions = [
    get_details,
    reset_password,
    cancel_appointment,
    schedule_local_intervention,
    schedule_distant_intervention,
    run_diagnostics,
    sign_up_to_option,
    cancel_option,
    get_customer_id
]

def get_partial_functions(customer_id):
    """
    Provide partial functions corresponding to a specific customer.
    These partial functions should be used when the customer is identified.
    """
    return [
        lambda s="": get_details(customer_id),
        lambda s="": reset_password(customer_id),
        lambda s="": cancel_appointment(customer_id),
        lambda s="": schedule_local_intervention(customer_id),
        lambda s="": schedule_distant_intervention(customer_id),
        lambda s="": run_diagnostics(customer_id),
        lambda s="": sign_up_to_option(f"{customer_id},{s}"),
        lambda s="": cancel_option(f"{customer_id},{s}"),
    ]

def get_tools(customer_id=None):
    """
    Provide either the general tools or the customer-specific tools.
    """
    tools = []
    if customer_id is None:
        for f in functions:
            tools.append(StructuredTool.from_function(f))
    else:
        partial_functions = get_partial_functions(str(customer_id))
        for i in range(len(partial_functions)):
            description = "\n".join(functions[i].__doc__.strip().split("\n    ")[:-1])
            # If partial functions are used, their description should be amended
            if functions[i].__name__ in ["sign_up_to_option", "cancel_option"]:
                description += "\nThe action input should be either 'TV' or 'Premium'"
            else:
                description += "\nThe action input should be the empty string"
            tools.append(StructuredTool.from_function(
                partial_functions[i],
                name=functions[i].__name__,
                description=description
            ))
    return tools

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
SYSTEM_PROMPT = """Respond to the human as helpfully and accurately as possible. You have access to the following tools:

{tools}

Use a json blob to specify a tool by providing an action key (tool name) and an action_input key (tool input).

Valid "action" values: "Final Answer" or {tool_names}

Provide only ONE action per $JSON_BLOB, as shown:

```
{{
  "action": $TOOL_NAME,
  "action_input": $INPUT
}}
```

Follow this format:

Question: input question to answer
Thought: consider previous and subsequent steps
Action:
```
$JSON_BLOB
```
Observation: action result
... (repeat Thought/Action/Observation N times)
Thought: I know what to respond
Action:
```
{{
  "action": "Final Answer",
  "action_input": "Final response to human"
}}

Begin! Reminder to ALWAYS respond with a valid json blob of a single action. Use tools if necessary. Respond directly if appropriate. Format is Action:```$JSON_BLOB```then Observation
"""

prompt = ChatPromptTemplate.from_messages([
    ("system", SYSTEM_PROMPT),
    MessagesPlaceholder(variable_name="chat_history", optional=True),
    HumanMessagePromptTemplate(prompt=PromptTemplate(input_variables=['agent_scratchpad', 'input'], template='{input}\n\n{agent_scratchpad}\n (reminder to respond in a JSON blob no matter what)'))
])

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
def serialize_intermediate_steps(l):
    """
    Serialize the intermediate steps returned by the agent.
    """
    return json.dumps([(x[0].tool, x[0].tool_input, x[0].log, x[1]) for x in l])

def process_request(request, customer_id=None):
    """
    Handle the customer's request.
    If the customer id is specified, only the corresponding customer can be affected by the agent.
    """
    mlflow_callback = MlflowCallbackHandler(
        experiment=EXPERIMENT_NAME,
        tags={"model": LLM_ID}
    )

    llm = DKUChatLLM(
        llm_id=LLM_ID,
        temperature=0,
        callbacks=[mlflow_callback]
    )
    
    chat_history = []
    if customer_id is not None:
        identity = get_details(str(customer_id)).split(",")[0]
        chat_history = [SystemMessage(content=f"This is a message received from {identity})")]

    tools = get_tools(customer_id=customer_id)
    agent = create_structured_chat_agent(llm, tools, prompt)
    agent_executor = AgentExecutor(
        agent=agent,
        tools=tools,
        return_intermediate_steps=True,
        handle_parsing_errors=True,
        callbacks=[mlflow_callback]
    )

    result = agent_executor.invoke({"input": request, "chat_history": chat_history})

    # List the actions and their details in two strings returned along with the answer
    actions = ""
    for j in range(len(result["intermediate_steps"])):
        tool = result["intermediate_steps"][j][0].tool
        tool_input = result["intermediate_steps"][j][0].tool_input
        thought = result["intermediate_steps"][j][0].log
        observation = result["intermediate_steps"][j][1]
        actions += f"{j+1}. {tool}({tool_input}) --> {observation}\n"

    return (
        result["output"],
        actions[:-1] if len(actions) > 0 else actions,
        serialize_intermediate_steps(result["intermediate_steps"])
    )

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
for i in range(len(df)):
    reply, actions, intermediate_steps = process_request(
        df.iloc[i].request,
        customer_id=df.iloc[i].customer_id
    )
    df.at[i, "draft_reply"] = reply
    df.at[i, "actions"] = actions
    df.at[i, "intermediate_steps"] = intermediate_steps

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
dataiku.Dataset("requests_processed_chat_model").write_with_schema(df)