# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# -*- coding: utf-8 -*-
import tempfile
import os
import dataiku
import torch
from transformers import AutoModelForCausalLM, AutoTokenizer, BitsAndBytesConfig
from peft import PeftModel
from transformers import StoppingCriteria, StoppingCriteriaList

folder = dataiku.Folder("74GhsK3Z")
df = dataiku.Dataset("sft_test_sampled").get_dataframe()

MODEL_NAME = "mistralai/Mistral-7B-Instruct-v0.2"
MODEL_REVISION = "c72e5d1908b1e2929ec8fc4c8820e9706af1f80f"
BATCH_SIZE = 64

# Use 4-bit quantization to the model to reduce its memory consumption
bnb_config = BitsAndBytesConfig(
    load_in_4bit=True,
    bnb_4bit_quant_type="nf4",
    bnb_4bit_compute_dtype=torch.float16,
)

# Load the base model
model = AutoModelForCausalLM.from_pretrained(
    MODEL_NAME,
    revision=MODEL_REVISION,
    quantization_config=bnb_config,
)

# Adjust the base model with the LoRA weights computed during fine-tuning
with tempfile.TemporaryDirectory() as temp_dir:
    for f in folder.list_paths_in_partition():
        with folder.get_download_stream(f) as stream:
            with open(os.path.join(temp_dir, os.path.basename(f)), "wb") as f2:
                f2.write(stream.read())
    ft_model = PeftModel.from_pretrained(model, temp_dir).eval()

tokenizer = AutoTokenizer.from_pretrained(MODEL_NAME, padding_side="left")
tokenizer.pad_token = tokenizer.eos_token

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Generate the answers with the fine-tuned model

i, results = 0, []
while i < len(df):
    batch = [f"[INST] Extract key details from: {s} [/INST]" for s in df.iloc[df.index[i:i+BATCH_SIZE]]["input"]]
    model_input = tokenizer(batch, return_tensors="pt", padding=True).to("cuda")
    num_samples, length_input = model_input.input_ids.shape
    with torch.no_grad():
        model_output = ft_model.generate(
            **model_input,
            max_new_tokens=512,
            pad_token_id=tokenizer.eos_token_id
        )
    for j in range(num_samples):
        results.append(tokenizer.decode(model_output[j, length_input:], skip_special_tokens=True).strip())
    i += BATCH_SIZE

df["llm_output"] = [x.split("}")[0] + "}" for x in results]

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
dataiku.Dataset("sft_predictions").write_with_schema(df)