import os
from datetime import datetime
import json
import dataiku
from dataiku.langchain.dku_llm import DKUChatLLM

from langchain_community.vectorstores import FAISS
from langchain.embeddings import HuggingFaceEmbeddings
from langchain.agents import AgentExecutor, create_structured_chat_agent
from langchain.agents.agent_toolkits import create_retriever_tool
from langchain_core.prompts import PromptTemplate, ChatPromptTemplate
from langchain_core.prompts import MessagesPlaceholder
from langchain_core.messages import AIMessage, HumanMessage
from langchain.pydantic_v1 import BaseModel, Field
from langchain.tools import StructuredTool

import dash
from dash import dcc, html, ctx
from dash.dependencies import Input, Output, State
import dash_bootstrap_components as dbc

WEBAPP_NAME = "qanda_chat_webapp" # Name of the app (logged when a conversation is flagged)
VERSION = dataiku.get_custom_variables()[WEBAPP_NAME] # Version of the app (logged when a conversation is flagged)
MODEL = "sentence-transformers/msmarco-distilbert-cos-v5"
LLM_ID = dataiku.get_custom_variables()["LLM_id"]

# Folder with the index of the input documents
folder = dataiku.Folder("4pbb7xKD")

# Folder to log conversations flagged by users
log_folder = dataiku.Folder("zhYE6r2h")

llm_provided = len(LLM_ID) > 0
llm = DKUChatLLM(
    llm_id=LLM_ID,
    temperature=0
)

ERROR_MESSAGE_MISSING_KEY = """
LLM connection missing. You need to add it as a user secret. Cf. this project's wiki.

Please note that this web app is not live on Dataiku’s public project gallery but you can test it by downloading the project and providing an LLM connection.
"""

SYSTEM_PROMPT = """Respond to the human as helpfully and accurately as possible. You have access to the following tools:
{tools}
Use a json blob to specify a tool by providing an action key (tool name) and an action_input key (tool input).
Valid "action" values: "Final Answer" or {tool_names}
Provide only ONE action per $JSON_BLOB, as shown:
```
{{

  "action": $TOOL_NAME,

  "action_input": $INPUT

}}
```
Follow this format:
Question: input question to answer
Thought: consider previous and subsequent steps
Action:
```
$JSON_BLOB
```
Observation: action result
... (repeat Thought/Action/Observation N times)
Thought: I know what to respond
Action:
```
{{

  "action": "Final Answer",

  "action_input": "Final response to human"

}}
```
Begin! Reminder to ALWAYS respond with a valid json blob of a single action. Use tools if necessary. Respond directly if appropriate. Format is Action:```$JSON_BLOB```then Observation
"""

if llm_provided:
    # Index for the input documents
    embeddings = HuggingFaceEmbeddings(
        model_name=MODEL,
        cache_folder=os.getenv("SENTENCE_TRANSFORMERS_HOME")
    )
    with tempfile.TemporaryDirectory() as temp_dir:
        for f in folder.list_paths_in_partition():
            with folder.get_download_stream(f) as stream:
                with open(os.path.join(temp_dir, os.path.basename(f)), "wb") as f2:
                    f2.write(stream.read())
        retriever = FAISS.load_local(temp_dir, embeddings, allow_dangerous_deserialization=True).as_retriever()
    
    tool = create_retriever_tool(
        retriever, 
        "search",
        "Search tool based on documents related to biodiversity. Always use this tool for biodiversity-related questions. The Action Input for this tool should be your search query.",
        document_prompt=PromptTemplate.from_template("{source}: {page_content}"),
        document_separator="\n\n-sep-\n\n"
    )
    tools = [tool]
    
    prompt = ChatPromptTemplate.from_messages([
        ("system", SYSTEM_PROMPT),
        MessagesPlaceholder(variable_name="chat_history"),
        ("human", "{input}\n\n\n\n{agent_scratchpad}")
    ])
    
    agent = create_structured_chat_agent(llm, tools, prompt)

def format_sources(s):
    """
    Format the sources so they can be displayed to the user.
    """
    escaped = s.replace("*", "\\*").replace("[", "\[").replace("]", "\]")
    docs = escaped.split("\n\n-sep-\n\n")
    for i in range(len(docs)):
        splitted = docs[i].split(":")
        docs[i] = f"**{splitted[0]}**: {':'.join(splitted[1:])}"
    return "\n\n---\n\n".join(docs)

def extract_sources(result):
    """
    Extract sources from the result of the conversational agent.
    """
    context = ""
    if "intermediate_steps" in result:
        for a in result["intermediate_steps"]:
            if a[0].tool == "search":
                context = format_sources(a[1])
    return context

def answer_question(history_questions, history_answers):
    """
    Answer the last question given all previous answers and questions.
    """
    
    chat_history = []
    for i in range(len(history_questions)-1):
        chat_history.append(HumanMessage(content=history_questions[i]))
        try:
            chat_history.append(AIMessage(content=history_answers[i][0]))
        except IndexError:
            pass
        
    agent_executor = AgentExecutor(
        agent=agent,
        tools=tools,
        verbose=True,
        handle_parsing_errors=True,
        return_intermediate_steps=True,
        early_stopping_method="force"
    )
    
    result = agent_executor({"input": history_questions[-1], "chat_history": chat_history, "agent_scratchpad": []})
    return (result["output"], extract_sources(result))

# Layout

STYLE_CONVERSATION = {
    "overflow-y": "auto",
    "display": "flex",
    "height": "calc(90vh - 50px)",
    "flex-direction": "column-reverse",
    "width": "100%"
}

STYLE_MESSAGE = {
    "max-width": "80%",
    "width": "max-content",
    "padding": "5px 10px",
    "border-radius": 10,
    "margin-bottom": 10,
}

STYLE_BUTTON_BAR = {
    "display": "flex",
    "justify-content": "center",
    "gap": "10px",
    "margin-top": "10px"
}

STYLE_ACCORDION = {
    "margin-top": "10px"
}

reset_icon = html.Span(html.I(className="bi bi-trash3"), style=dict(paddingRight="5px"))
flag_icon = html.Span(html.I(className="bi bi-flag"), style=dict(paddingRight="5px"))

button_bar = html.Div(
    [
        dbc.Button(
            html.Span([reset_icon, 'Reset conversation']),
            id='reset-btn',
            title='Delete all previous messages'
        ),
        dbc.Button(
            html.Span([flag_icon, 'Flag conversation']),
            id='flag-btn',
            title='Flag the conversation, e.g. in case of inappropriate or erroneous replies'
        ),
    ],
    style=STYLE_BUTTON_BAR
)

send_icon = html.Span(html.I(className="bi bi-send"))

question_bar = html.Div(
    [
        dbc.InputGroup(
            [
                dbc.Input(id='query', value='', type='text', minLength=0),
                dbc.Button(send_icon, id='send-btn', title='Get an answer')
            ],
        ),
        button_bar
    ]
)

conversation = html.Div(
    html.Div(id="conversation"),
    style=STYLE_CONVERSATION
)

spinning_wheel = dbc.Spinner(
    dcc.Markdown(
        id='spinning_wheel',
        style={
            "height": "20px",
            "margin-bottom": "20px"
        }
    ),
    color="primary"
)

app.title = "Chatbot"
app.config.external_stylesheets = [
    dbc.themes.ZEPHYR,
    dbc.icons.BOOTSTRAP
]

app.layout = html.Div(
    [  
        conversation,
        spinning_wheel,
        question_bar,
        dcc.Store(id="logged"),
        dcc.Store(id='history_questions', storage_type='memory', data=[]),
        dcc.Store(id='history_answers', storage_type='memory', data=[]),
    ],
    style={
        "margin": "auto",
        "text-align": "left",
        "max-width": "800px"
    }
)

# Callbacks

def textbox(text, box="AI", accordion=""):
    """
    Create the text box corresponding to a message.
    """
    style = dict(STYLE_MESSAGE)
    if box == "user":
        style["margin-left"] = "auto"
        style["margin-right"] = 0
        color, inverse = "primary", True
    else:
        style["margin-left"] = 0
        style["margin-right"] = "auto"
        color, inverse = "light", False
    content = [html.Div(text)]
    if len(accordion) > 0:
        content.append(
            dbc.Accordion(
                [
                    dbc.AccordionItem(
                        dcc.Markdown(accordion),
                        title="Sources",
                    )
                ],
                start_collapsed=True,
                style=STYLE_ACCORDION
            )
        )
    return dbc.Card(
        html.Div(content),
        style=style,
        body=True,
        color=color,
        inverse=inverse
    )

@app.callback(
    Output("conversation", "children"),
    Input("history_questions", "data"),
    Input("history_answers", "data"),
    prevent_initial_call=True
)
def update_display(history_questions, history_answers):
    """
    Display the messages of the conversation.
    """
    result = []
    for i in range(len(history_questions)):
        result.append(textbox(history_questions[i], box="user"))
        try:
            answer = history_answers[i]
            result.append(textbox(answer[0], accordion=answer[1], box="AI"))
        except IndexError:
            pass
    return result

@app.callback(
    Output('history_questions', 'data'),
    Output('query', 'value'),
    Input('reset-btn', 'n_clicks'),
    Input('send-btn', 'n_clicks'),
    Input('query', 'n_submit'),
    State('query', 'value'),
    State('history_questions', 'data'),
    State('history_answers', 'data'),
    prevent_initial_call=True
)
def receive_query(reset, n_clicks, n_submit, query, history_questions, history_answers):
    """
    Receive the new query from the user.
    """
    if not llm_provided:
        return [ERROR_MESSAGE_MISSING_KEY], ""
    if ctx.triggered_id == "reset-btn":
        return [], ""
    if len(query) > 0 and len(history_questions) == len(history_answers):
        history_questions.append(query)
        return history_questions, ""
    else:
        return history_questions, query

@app.callback(
    Output('history_answers', 'data'),
    Output('spinning_wheel', 'children'),
    Input('history_questions', 'data'),
    State('history_answers', 'data'),
    prevent_initial_call=True
)
def get_answer(history_questions, history_answers):
    """
    Receive the answer from the LLM.
    """
    if not llm_provided:
        return [], ""
    if len(history_questions) == 0:
        return [], ''
    if len(history_questions) > len(history_answers):
        answer = answer_question(history_questions, history_answers)
        history_answers.append(answer)
    return list(history_answers), ''

@app.callback(
    Output('logged', 'data'),
    Input('flag-btn', 'n_clicks'),
    Input('history_questions', 'data'),
    Input('history_answers', 'data'),
)
def log_conversation(n_clicks, history_questions, history_answers):
    """
    Log the current conversation.
    """
    if len(history_answers) > 0 and ctx.triggered_id == "flag-btn":
        path = f"/{hash(tuple(history_questions + history_answers))}.json"
        with log_folder.get_writer(path) as w:
            w.write(bytes(json.dumps({
                "history_questions": history_questions,
                "history_answers": history_answers,
                "timestamp": str(datetime.now()),
                "webapp": WEBAPP_NAME,
                "version": VERSION
            }), "utf-8"))
        return path
    else:
        return ""

@app.callback(
    Output('reset-btn', 'disabled'),
    Output('flag-btn', 'disabled'),
    Output('send-btn', 'disabled'),
    Input('history_questions', 'data'),
    Input('history_answers', 'data'),
    Input('logged', 'data'),
)
def disable_buttons(history_questions, history_answers, flagged):
    """
    Disable buttons when appropriate.
    """
    if len(history_answers) == 0:
        disable_reset = True
        disable_flag = True
    else:       
        disable_reset = False
        disable_flag = True if flagged != "" else False
    disable_send = True if len(history_questions) > len(history_answers) else False
    return disable_reset, disable_flag, disable_send