import os
from datetime import datetime
import json
import dataiku
from dataiku.langchain.dku_llm import DKUChatLLM

from langchain.agents import AgentExecutor, create_structured_chat_agent
from langchain_core.messages import AIMessage, HumanMessage, SystemMessage
from langchain_core.prompts import ChatPromptTemplate, PromptTemplate, HumanMessagePromptTemplate
from langchain_core.prompts import MessagesPlaceholder
from langchain.tools import StructuredTool

from langchain.tools import WikipediaQueryRun
from langchain_community.utilities import WikipediaAPIWrapper

import dash
from dash import dcc, html, ctx
from dash.dependencies import Input, Output, State
import dash_bootstrap_components as dbc

WEBAPP_NAME = "agent_chat_webapp" # Name of the app (logged when a conversation is flagged)
VERSION = "1.0" # Version of the app (logged when a conversation is flagged)
LLM_ID = dataiku.get_custom_variables()["LLM_id"]

# Folder to log conversations flagged by users
log_folder = dataiku.Folder("zhYE6r2h")

llm_provided = len(LLM_ID) > 0
llm = DKUChatLLM(
    llm_id=LLM_ID,
    temperature=0
)

ERROR_MESSAGE_MISSING_KEY = """
LLM connection missing. You need to add it as a project variable. Cf. this project's wiki.

Please note that this web app is not live on Dataiku’s public project gallery but you can test it by downloading the project and providing an LLM connection.
"""

def sum_numbers(x: int, y: int) -> int:
    """
    Return the sum of the two input numbers.
    """
    return x + y


sum_tool = StructuredTool.from_function(
    func=sum_numbers
)

tools = [sum_tool, WikipediaQueryRun(api_wrapper=WikipediaAPIWrapper())]

SYSTEM_PROMPT = """Respond to the human as helpfully and accurately as possible. You have access to the following tools:

{tools}

Use a json blob to specify a tool by providing an action key (tool name) and an action_input key (tool input).

Valid "action" values: "Final Answer" or {tool_names}

Provide only ONE action per $JSON_BLOB, as shown:

```
{{
  "action": $TOOL_NAME,
  "action_input": $INPUT
}}
```

Follow this format:

Question: input question to answer
Thought: consider previous and subsequent steps
Action:
```
$JSON_BLOB
```
Observation: action result
... (repeat Thought/Action/Observation N times)
Thought: I know what to respond
Action:
```
{{
  "action": "Final Answer",
  "action_input": "Final response to human"
}}

Begin! Reminder to ALWAYS respond with a valid json blob of a single action. Use tools if necessary. Respond directly if appropriate. Format is Action:```$JSON_BLOB```then Observation
"""

prompt = ChatPromptTemplate.from_messages([
    ("system", SYSTEM_PROMPT),
    MessagesPlaceholder(variable_name="chat_history", optional=True),
    HumanMessagePromptTemplate(prompt=PromptTemplate(input_variables=['agent_scratchpad', 'input'], template='{input}\n\n{agent_scratchpad}\n (reminder to respond in a JSON blob no matter what)'))
])

agent = create_structured_chat_agent(llm, tools, prompt)
agent_executor = AgentExecutor(
    agent=agent,
    tools=tools,
    verbose=True,
    return_intermediate_steps=True,
    handle_parsing_errors=True,
    max_iterations=3,
    early_stopping_method="force"
)

def answer_question(history_questions, history_answers):
    """
    Answer the last question given all previous answers and questions.
    """
    
    chat_history = []
    for i in range(len(history_questions)-1):
        chat_history.append(HumanMessage(content=history_questions[i]))
        try:
            chat_history.append(AIMessage(content=history_answers[i][0]))
        except IndexError:
            pass
    result = agent_executor.invoke({"input": history_questions[-1], "chat_history": chat_history})
    intermediate_steps = ""
    for x in result["intermediate_steps"]:
        intermediate_steps += f"{x[0].tool}({x[0].tool_input}) --> {x[1]}\n"
    return (result["output"], intermediate_steps.strip())

# Layout

STYLE_CONVERSATION = {
    "overflow-y": "auto",
    "display": "flex",
    "height": "calc(90vh - 50px)",
    "flex-direction": "column-reverse",
    "width": "100%"
}

STYLE_MESSAGE = {
    "max-width": "80%",
    "width": "max-content",
    "padding": "5px 10px",
    "border-radius": 10,
    "margin-bottom": 10,
}

STYLE_BUTTON_BAR = {
    "display": "flex",
    "justify-content": "center",
    "gap": "10px",
    "margin-top": "10px"
}

STYLE_ACCORDION = {
    "margin-top": "10px"
}

reset_icon = html.Span(html.I(className="bi bi-trash3"), style=dict(paddingRight="5px"))
flag_icon = html.Span(html.I(className="bi bi-flag"), style=dict(paddingRight="5px"))

button_bar = html.Div(
    [
        dbc.Button(
            html.Span([reset_icon, 'Reset conversation']),
            id='reset-btn',
            title='Delete all previous messages'
        ),
        dbc.Button(
            html.Span([flag_icon, 'Flag conversation']),
            id='flag-btn',
            title='Flag the conversation, e.g. in case of inappropriate or erroneous replies'
        ),
    ],
    style=STYLE_BUTTON_BAR
)

send_icon = html.Span(html.I(className="bi bi-send"))

question_bar = html.Div(
    [
        dbc.InputGroup(
            [
                dbc.Input(id='query', value='', type='text', minLength=0),
                dbc.Button(send_icon, id='send-btn', title='Get an answer')
            ],
        ),
        button_bar
    ]
)

conversation = html.Div(
    html.Div(id="conversation"),
    style=STYLE_CONVERSATION
)

spinning_wheel = dbc.Spinner(
    dcc.Markdown(
        id='spinning_wheel',
        style={
            "height": "20px",
            "margin-bottom": "20px"
        }
    ),
    color="primary"
)

app.title = "Chatbot"
app.config.external_stylesheets = [
    dbc.themes.ZEPHYR,
    dbc.icons.BOOTSTRAP
]

app.layout = html.Div(
    [  
        conversation,
        spinning_wheel,
        question_bar,
        dcc.Store(id="logged"),
        dcc.Store(id='history_questions', storage_type='memory', data=[]),
        dcc.Store(id='history_answers', storage_type='memory', data=[]),
    ],
    style={
        "margin": "auto",
        "text-align": "left",
        "max-width": "800px"
    }
)

# Callbacks

def textbox(text, box="AI", accordion=""):
    """
    Create the text box corresponding to a message.
    """
    style = dict(STYLE_MESSAGE)
    if box == "user":
        style["margin-left"] = "auto"
        style["margin-right"] = 0
        color, inverse = "primary", True
    else:
        style["margin-left"] = 0
        style["margin-right"] = "auto"
        color, inverse = "light", False
    content = [html.Div(text)]
    if len(accordion) > 0:
        content.append(
            dbc.Accordion(
                [
                    dbc.AccordionItem(
                        dcc.Markdown(accordion),
                        title="Actions taken",
                    )
                ],
                start_collapsed=True,
                style=STYLE_ACCORDION
            )
        )
    return dbc.Card(
        html.Div(content),
        style=style,
        body=True,
        color=color,
        inverse=inverse
    )

@app.callback(
    Output("conversation", "children"),
    Input("history_questions", "data"),
    Input("history_answers", "data"),
    prevent_initial_call=True
)
def update_display(history_questions, history_answers):
    """
    Display the messages of the conversation.
    """
    result = []
    for i in range(len(history_questions)):
        result.append(textbox(history_questions[i], box="user"))
        try:
            answer = history_answers[i]
            result.append(textbox(answer[0], accordion=answer[1], box="AI"))
        except IndexError:
            pass
    return result

@app.callback(
    Output('history_questions', 'data'),
    Output('query', 'value'),
    Input('reset-btn', 'n_clicks'),
    Input('send-btn', 'n_clicks'),
    Input('query', 'n_submit'),
    State('query', 'value'),
    State('history_questions', 'data'),
    State('history_answers', 'data'),
    prevent_initial_call=True
)
def receive_query(reset, n_clicks, n_submit, query, history_questions, history_answers):
    """
    Receive the new query from the user.
    """
    if not llm_provided:
        return [ERROR_MESSAGE_MISSING_KEY], ""
    if ctx.triggered_id == "reset-btn":
        return [], ""
    if len(query) > 0 and len(history_questions) == len(history_answers):
        history_questions.append(query)
        return history_questions, ""
    else:
        return history_questions, query

@app.callback(
    Output('history_answers', 'data'),
    Output('spinning_wheel', 'children'),
    Input('history_questions', 'data'),
    State('history_answers', 'data'),
    prevent_initial_call=True
)
def get_answer(history_questions, history_answers):
    """
    Receive the answer from the LLM.
    """
    if not llm_provided:
        return [], ""
    if len(history_questions) == 0:
        return [], ''
    if len(history_questions) > len(history_answers):
        result = answer_question(history_questions, history_answers)
        history_answers.append(result)
    return list(history_answers), ''

@app.callback(
    Output('logged', 'data'),
    Input('flag-btn', 'n_clicks'),
    Input('history_questions', 'data'),
    Input('history_answers', 'data'),
)
def log_conversation(n_clicks, history_questions, history_answers):
    """
    Log the current conversation.
    """
    if len(history_answers) > 0 and ctx.triggered_id == "flag-btn":
        path = f"/{hash(tuple(history_questions + [x[0] + x[1] for x in history_answers]))}.json"
        with log_folder.get_writer(path) as w:
            w.write(bytes(json.dumps({
                "history_questions": history_questions,
                "history_answers": history_answers,
                "timestamp": str(datetime.now()),
                "webapp": WEBAPP_NAME,
                "version": VERSION
            }), "utf-8"))
        return path
    else:
        return ""

@app.callback(
    Output('reset-btn', 'disabled'),
    Output('flag-btn', 'disabled'),
    Output('send-btn', 'disabled'),
    Input('history_questions', 'data'),
    Input('history_answers', 'data'),
    Input('logged', 'data'),
)
def disable_buttons(history_questions, history_answers, flagged):
    """
    Disable buttons when appropriate.
    """
    if len(history_answers) == 0:
        disable_reset = True
        disable_flag = True
    else:       
        disable_reset = False
        disable_flag = True if flagged != "" else False
    disable_send = True if len(history_questions) > len(history_answers) else False
    return disable_reset, disable_flag, disable_send