# Overview

This _**Dataiku LLM Starter Kit**_ shows how Large Language Models (LLMs) like ChatGPT or Mistral can be used in Dataiku. It illustrates the notions presented in the ["Introduction to Large Language Models With Dataiku" guidebook](https://content.dataiku.com/llms-dataiku/dataiku-llm-starter-kit).

**You are invited to read the [guidebook](https://content.dataiku.com/llms-dataiku/dataiku-llm-starter-kit) and learn about the notion of [LLM Mesh](https://doc.dataiku.com/dss/latest/generative-ai/introduction.html) before exploring this project**.

The project shows how to:
- Perform various Natural Language Processing (NLP) tasks with **visual recipes**;
- Call LLMs with Python **code recipes** to better control the text generation process;
- Implement a **question answering system based on documents** in various formats;
- Augment LLMs with **external tools** that enable them to act in the real world;
- Increase the likelihood that the text generated by an LLM complies with an **expected format**;
- **Detect potentially harmful content** generated by an LLM;
- **Fine-tune an open source LLM**;
- Implement a **conversational agent**.

The project can be [downloaded](https://downloads.dataiku.com/public/dss-samples/EX_LLM_STARTER_KIT/) and the [last section](#next-develop-your-own-llm-use-cases-1) provides instructions to reuse it with your own datasets.

# Walkthrough

## Performing natural language processing tasks without code

From Dataiku version 12.3 onwards, using visual recipes is the simplest way to leverage LLMs in Dataiku.

In the [1. Basic use](flow_zone:JSx6cEw) Flow zone, we illustrate such recipes with a [small sample](dataset:product_reviews) of the [Amazon Reviews dataset](https://nijianmo.github.io/amazon/index.html). This dataset includes 21 reviews distributed in 7 product categories ("Cell Phones and Accessories", "Digital Music", "Electronics", "Industrial and Scientific", "Office Products", "Software", "Toys and Games") and classified as "positive", "negative" or "neutral".

For example, a review categorized as "Cell Phones and Accessories" and "positive" is:
```
Can't say enough good things. Fits snugly and provides great all round protection. Good grip but slides easily into pocket. Really adds to the quality feel of the Moto G 2014. I had the red and it looks sharp. Highly recommended.
```

### Text classification

If we first want to classify the reviews in various categories, we can simply create a ["Classify text" recipe](recipe:compute_classification2), list the targeted classes and optionally specify some few-shot examples. In this way, we can perform **zero-shot or few-shot in-context learning** as described in the ["Introduction to Large Language Models With Dataiku" guidebook](https://content.dataiku.com/llms-dataiku/dataiku-llm-starter-kit). It is also possible to request row-level explanations for the predicted classes.

The [resulting dataset](dataset:product_reviews_classified) includes several columns in addition to those of the input dataset: one for the predictions, one for the explanations (if requested), one for the LLM raw response and for potential LLM error messages.
![classification_v12.3.png](85suKFenVGEt)

### Text summarization

Similarly, a ["Summarize" recipe](recipe:compute_product_reviews_summarized) can be used to summarize the product reviews with an optional limit on the number of words or sentences.
![summarization_v12.3.png](D0H9zMf2jBvJ)

### Text generation

The two previous recipes are specific to certain tasks. More generally, the **[Prompt recipe](https://knowledge.dataiku.com/latest/ml-analytics/gen-ai/concept-prompt-studio.html)** allows to use any prompt. This enables users to easily perform a wide range of NLP tasks. The prompt template, the few-shot examples and various settings can be specified directly through the Prompt recipe or via Prompt Studios. We use this recipe to accomplish a [translation task](recipe:compute_product_reviews_translated) and a [structured information extraction task](recipe:compute_product_reviews_extracted).
![generation_v12.3.png](5IoS2dHBavHU)

## Better steering the text generation process with code recipes

The visual LLM recipes enable Dataiku users to easily accomplish natural language processing tasks. As shown in the [2. Basic use (with code)](flow_zone:default) Flow zone, we can go a step further with **code recipes**.

### LLM connection

If an [LLM connection](https://doc.dataiku.com/dss/latest/generative-ai/llm-connections.html) has been configured by the adminitrator of the Dataiku instance, we can use it in both visual and code recipes. In this way, we can take advantage of all the benefits of the [LLM Mesh](https://blog.dataiku.com/llm-mesh): decoupling application from service layer, enforcing a secure gateway, controlling cost and performance...). For instance, we show how to perform [few-shot sentiment analysis](recipe:compute_sentiment_llm_mesh) on the same [dataset](dataset:product_reviews) as before.

### OpenAI

Directly querying third-party APIs can offer some advantages (but does not bring the benefits of the LLM Mesh). In two Python code recipes, we perform the same sentiment analysis task with the `openai` library:
- ... and the `gpt-3.5-turbo` model, [using](recipe:compute_sentiment) the OpenAI chat completion endpoint;
- ... and the `gpt-3.5-turbo-instruct` model, [using](recipe:compute_sentiment_probabilities) the OpenAI text completion endpoint.

In addition to predicting the class with few-shot demonstrations, the recipes:
- Restrict the set of tokens generated so that **only the expected class labels can be predicted** (here: "positive", "neutral", "negative");
- Use the probability provided by the language model to compute a **confidence score**.

The recipes work only when the class labels are one-token long. The approach remains valid for more than one token but it is harder to implement. In this case, it is better to use more advanced tools like [LMQL](https://lmql.ai/), a “a programming language for LLMs”.

### Hugging Face model

We finally [replace](recipe:compute_sentiment_mistral) these proprietary models with a **locally-hosted open source LLM**. This gives more control on the text generation process and does not require sending data to a third party. At the time of writing, open source models are however less capable according to [several](https://crfm.stanford.edu/helm/latest/) [benchmarks](https://arena.lmsys.org/), they may not be allowed for commercial applications and using them can be challenging from an infrastructure perspective. Please also note that some of the most popular Hugging Face models can also be hosted through an [LLM connection](https://doc.dataiku.com/dss/latest/generative-ai/llm-connections.html).

## Answering questions with documents

In the [3. Retrieval-based question answering - visual](flow_zone:nS8CTF4) and [4. Retrieval-based question answering - code](flow_zone:wwVg0ys) Flow zones, we create a question-answering system based on a specific collection of documents. More specifically, we implement a ***retrieval-augmented generation* pipeline** which includes the following steps:
1. Receiving the question from the user;
2. Retrieving (from the collection of documents) the passages most semantically similar to the question;
3. Incorporating both the question and the selected passages in a question-answering prompt ;
4. Querying a LLM with this prompt and receiving the answer.

You can find more details about the retrieval-augmented generation pipeline in the ["Introduction to Large Language Models With Dataiku" guidebook](https://content.dataiku.com/llms-dataiku/dataiku-llm-starter-kit) or this [blog post](https://medium.com/data-from-the-trenches/digging-through-the-minutiae-query-your-documents-with-gpt-3-c4600635a55).

The pipeline in the [3. Retrieval-based question answering - visual](flow_zone:nS8CTF4) Flow zone is based on visual features available from Dataiku version 12.3 onwards while the pipelline in the [4. Retrieval-based question answering - code](flow_zone:wwVg0ys) Flow zone is based only on code recipes and leverages [LangChain](https://python.langchain.com/en/latest/), a rich and fast-evolving toolkit to create LLM-powered applications.

### Loading, splitting and vectorizing documents

Before answering questions, we need to **preprocess the documents** once and for all (at least until the next update of the documents) to enable the retrieval of relevant passages at Step 2 above.

In our example, we want to answer questions with [4 documents](managed_folder:TyR7HVoz) related to biodiversity: 2 .pdf documents, 1 .docx document and 1 .html page. We first:
- Load the content of the documents;
- Split the documents' content in chunks;
- Compute the embeddings of these chunks with an open source semantic similarity model;
- Create an index of these embeddings to be able to quickly retrieve them;
- Save this index in a vector store.

This happens:
- in the [3. Retrieval-based question answering - visual](flow_zone:nS8CTF4) Flow zone, through:
  - a first [visual recipe](recipe:compute_documents_extracted) of the [Text Extraction and OCR plugin](https://www.dataiku.com/product/plugins/tesseract-ocr/) to load the documents' content;
  - a [Prepare recipe](recipe:compute_documents_extracted_prepared) to clean the documents' content, enrich the extracts with relevant metadata and create a meaningful extract id (for example,  `wikipedia_biodiversity.pdf - Page 10`) to better identify the sources of an answer. This step is optional;
  - a second [visual recipe](recipe:compute_knowledge) for all the other steps.
- in the [4. Retrieval-based question answering - code](flow_zone:wwVg0ys) Flow zone through a [Python recipe](recipe:compute_4pbb7xKD).

### Generating answers and sources

With the document indexed, it is then straightforward to implement the retrieval-augmented generation pipeline described above. 

In the [3. Retrieval-based question answering - visual](flow_zone:nS8CTF4) Flow zone, we use a [Prompt recipe](recipe:compute_answers2) and leverage an augmented LLM connection to get [answers](dataset:answers) for a [set of questions](dataset:questions).

In the [4. Retrieval-based question answering - code](flow_zone:wwVg0ys) Flow zone, we use a [Python recipe](recipe:compute_answers) to generate [answers](dataset:answers2) for the same set of questions. More precisely, we take advantage of the "QA with sources" chain provided in Langchain. In the recipe, this answer is post-processed to include the content of the chunks mentioned in the sources when their complete identifier (here, "wikipedia_biodiversity.pdf, page 4 (1/2)" for example) is provided. This is not always the case because the LLM may just mention "wikipedia_biodiversity.pdf" or "wikipedia_biodiversity.pdf (multiple pages)".

A very simple [web app](web_app:gfMP9Yq) reusing the code of the previous [Python recipe](recipe:compute_answers) allows users to get answers and provide feedback (😀 or 🙁) on the answers (in which case, the query, the answer and the feedback are [logged](managed_folder:zhYE6r2h). **Please note that you need to download the project, import it on your own Dataiku instance and provide an OpenAI key to test the web app. It cannot be used on Dataiku's public gallery**. Two alternative web apps are available: [one using a conversational interface](web_app:1yaSXSp) and a minimalistic [one using only a Dataiku LLM connection](web_app:qpPlJBw), without LangChain.
![Question answering web app](gASJVkUuWqtZ)

Moreover, another [web app](web_app:MNIEqRT) enables you to interactively search the [knowledge bank](retrievable_knowledge:iXhIIXV5) used in the [3. Retrieval-based question answering - visual](flow_zone:nS8CTF4) Flow zone. It can help debug a question-answering pipeline and it requires Dataiku v12.4.1 or above. You can learn more about semantic search through this [blog post](https://medium.com/data-from-the-trenches/semantic-search-an-overlooked-nlp-superpower-b67c4b1b119a) and this [Dataiku example project](https://gallery.dataiku.com/projects/EX_SEMANTICSEARCH/).
![semantic_search.png](CC3AOEIthgzV)

### Automatically comparing generated answers and ground-truth answers

Evaluating a question answering system is challenging because various valid answers can be phrased differently. Simply comparing word for word generated answers and ground-truth answers would most likely underestimate the proportion of correct answers. We show three methods to overcome this problem and **perform automated (but potentially noisy) evaluations**:
- We simply ask an LLM to compare a generated answer and a ground-truth answer and determine whether they are equivalent;
- Following the [BERTScore: Evaluating Text Generation with BERT](https://arxiv.org/abs/1904.09675) paper, we compute an automatic [evaluation metric](https://huggingface.co/spaces/evaluate-metric/bertscore) based on the embeddings of the tokens in both the generated answer and the ground-truth answer;
- We use [MLflow LLM Evaluate](https://mlflow.org/docs/latest/llms/llm-evaluate/index.html) to compute various LLM-as-a-judge metrics thanks to an LLM.

This evaluation takes place in two similar Python recipes ([one](recipe:compute_answers_assessed) in the [3. Retrieval-based question answering - visual](flow_zone:nS8CTF4) Flow zone and [one](recipe:compute_answers_evaluated) in the [4. Retrieval-based question answering - code](flow_zone:wwVg0ys) Flow zone) and the evaluation results are recorded in this [evaluation store](model_evaluation_store:3E3N8G5w) and this [one](model_evaluation_store:jNuKpcAc).

## Augmenting an LLM with external tools

In the [5. LLMs combined with tools](flow_zone:Gzn9KXV) Flow zone, we create an ***agent***, ie. an LLM augmented with ***tools*** so that it can act in the real world, access external knowledge sources, or better perform certain tasks. For this fictitious example, we imagine the case of an Internet provider handling [customer requests](dataset:requests) related to Internet connection troubles, password reinitialization, options...

### Defining tools

In this context, tools are simply programs that take some text as input and provide or summarize their results as text. Here, we consider 9 tools (included in the `fictitious_tools.py` file in the project's libraries):
- `get_customer_id` retrieves the customer ID given his or her name;
- `get_details` retrieves the customer data of a specific customer;
- `reset_password` reinitializes the password of a specific customer;
- `schedule_local_intervention` sends an email allowing a specific customer to schedule an appointment for a technician's visit;
- `schedule_distant_intervention` sends an email allowing a specific customer to schedule an appointment for a phone discussion with a technician;
- `cancel_appointment` cancels an upcoming appointment;
- `run_diagnostics` determines whether an Internet connection problem requires the local or distant intervention of a technician;
- `sign_up_to_option` activates the Premium option or the TV option for a specific customer;
- `cancel_option` deactivates the Premium option or the TV option for a specific customer.

For example, the `reset_password` tool is the following Python function:
```
def reset_password(customer_id):
    """
    Reset the password of a customer.
    Use this if the customer mentions problems with his or her password.
    The input should be a single integer representing the customer's id. E.g. '123'.
    """
    try:
        row = customers_df.loc[extract_integer(customer_id)]
        return f"Email sent so that customer #{customer_id} can define a new password. No further action is needed to reinitialize the password."
    except (KeyError, ValueError):
        return "Unknown customer id"
```
Since our example is fictitious, this function is "fake", in the sense that it does not really reset a password. It just confirms that it did. Of course, tools can be arbitrarily complex functions: they may call API endpoints, they may leverage a predictive model... They may even be another LLM augmented with other tools. What matters here from the perspective of the LLM is that **the function is described in a docstring and returns an intelligible string**.

### Implementing the ReAct approach

Once the tools are defined, we can [create](recipe:recipe_from_notebook_vivien_tranthiens_Python_notebook) an agent based on the **ReAct** approach (you can read more about ReAct in the ["Introduction to Large Language Models with Dataiku" guidebook](https://content.dataiku.com/llms-dataiku/dataiku-llm-starter-kit) or in the original [research paper](https://arxiv.org/abs/2210.03629)).

This involves writing the prompt template and implementing all the machinery to call the LLM, parse its answers and execute the actions requested by the LLM with the tools. Fortunately, **LangChain** makes this straightforward.

Our code allows for two variants. In the default situation, the agent can potentially affect all customers. This creates the risk that handling a request inadequately impacts another customer. In the alternative situation, the customer is assumed to be known. We can then replace the generic tools by **restricted tools** specific to this customer. Given that agents can make mistakes and take inadequate actions, it is obviously better to use restricted tools whenever possible.

Given some fictitious [customer data](dataset:customers), we use this new agent to [process](recipe:recipe_from_notebook_vivien_tranthiens_Python_notebook) [examples of requests](dataset:requests) using the restricted tools. The results are satisfying. For example, with the following request:

```
Hi. I have a problem with my Internet connection. Can you please solve this? Many thanks!
```
... we get the following reply:
```
I have scheduled a technician to visit you and solve the Internet connection problem. Please let me know if you need any further assistance.
```
... with these actions:
```
run_diagnostics()
schedule_local_intervention()
```
Just with the description of the tools, the agent correctly inferred that it should run some diagnostics as soon as an Internet connection problem is reported. It then appropriately chose to organize the visit of a technician. The draft reply is also appropriate and well written.

The course of action and the draft reply of the agent were adequate for 8 of the [10 examples of requests](dataset:requests). In the following case for example, the agent resets the password but does not address the internet connection problem.
```
Hi. My Internet access doesn’t work anymore. Moreover, I can’t remember what my password is. Please help!
```
This shows that **agents can make mistakes and should be rigorously developed and evaluated** (cf. the ["Introduction to Large Language Models with Dataiku" guidebook](https://content.dataiku.com/llms-dataiku/dataiku-llm-starter-kit) for practical recommendations on making agents more robust).

You can test the agent through a [web app](web_app:e5X6LOZ), with either the general tools (when the customer is not identified) or the restricted tools specific to a certain customer. **Please note that you need to download the project, import it on your own Dataiku instance and provide an OpenAI API key to test the web app. It cannot be used on Dataiku's public gallery.**

![agent.png](lCWoxTemTO6f)

### Using a chat model to power the agent

With `langchain`, both text completion models and chat completion models can power an agent. An agent similar as the previous one is implemented in a [code recipe](recipe:compute_requests_processed_chat_model). One advantage of using a chat model is that the agent can be used in a conversational context, as demonstrated in a [web app](web_app:GSkVIyJ) that can easily be reused for an agent use case.

![conversational_interface_agent.png](95RQQzvPR8V8)

### Leveraging the "function calling" OpenAI feature

In June 2023, OpenAI announced new variants of `gpt-3.5-turbo` and `gpt-4` that can be given descriptions of functions and, when relevant, request the use of a function and provide the corresponding arguments in a JSON object. This "[function calling](https://openai.com/blog/function-calling-and-other-api-updates)" feature is of course **particularly useful for LLM agents** and has since been incorporated in LangChain. We provide an [alternative version](recipe:compute_requests_processed_openai_function_call) of the Python recipe leveraging this feature.

### Evaluating the answers of an agent

Evaluating an agent is challenging because we may need to assess both its final answers and its trajectories, ie. all the actions it took along the way. For both the final answers and the trajectories, there may be several distinct but equally valid choices. The LLM Starter Kit illustrates **three general evaluation methods for agents**. The table below summarizes their required inputs, their outputs and the Python packages leveraged in the evaluation recipe.

| Evaluation method | Required inputs | Outputs | Python package |
|:---:|:---:|:---:|:---:|
| **LLM-as-a-judge**<br>(without ground truth) | - Request<br>- Reply<br>- Intermediate steps | - Score<br>- Justification | `langchain` |
| **LLM-as-a-judge**<br>(with ground truth) | - Request<br> - Reply<br>- Ground truth reply | - Score<br>- Justification | `mlflow` |
| **Conformance checking**<br>(process mining technique) | - Intermediate steps<br>- Model for the ground truth trajectory | - Trajectory fit score<br>- Aligned trajectories | `pm4py` |

The **first evaluation method** simply asks an LLM to score and comment the reply and the trajectory on the basis of the request. The LLM is not provided a ground truth answer or a ground truth trajectory, which makes this method convenient but limits its accuracy.

The **second evaluation method** is the same but the LLM is given access to a ground truth reply but not the intermediate steps. This may be relevant when only the final answer matters and when inadequate actions are inexpensive and have no side effects.

The  **third evaluation method** uses the trajectory of the agent and a *model* of the ground truth trajectory, ie. the list of required actions and the potential constraints on their execution. With these pieces of information, we can use a [process mining technique](https://en.wikipedia.org/wiki/Conformance_checking) called **conformance checking** to score the trajectory of the agent and align, to the extent possible, this trajectory and an ideal trajectory. This allows to detect inadequate trajectories and, within these trajectories, which actions are missing or superfluous.

For example, if the user's request is "I would like to subscribe to the TV and Premium options", a model for the ground truth trajectory could be:
```
PARALLEL:
  - sign_up_to_option({"s"="TV"})
  - sign_up_to_option({"s"="Premium"})
```
... which means that the actions `sign_up_to_option({"s": "TV"})` and ``sign_up_to_option({"s": "Premium"})`` should both take place, in any order. If the trajectory of the agent is only:
```
sign_up_to_option({"s": "TV"})
```
... then the trajectory fit will be 0.667 (a score less than 1 denotes an incorrect trajectory) and the aligned trajectories will be:

| Trajectory of the agent | Aligned ground-truth trajectory |
|:---:|:---:|
| - | sign_up_to_option({"s": "Premium"}) |
| sign_up_to_option({"s": "TV"}) | sign_up_to_option({"s": "TV"}) |

These aligned trajectories show that the `sign_up_to_option({"s": "Premium"}) ` action is missing from the trajectory of the agent.

The first and second evaluation methods are implemented in a [code recipe](recipe:compute_agent_answers_evaluated) of the project. In contrast, the third evaluation method only appears in a [companion project](https://gallery.dataiku.com/projects/EX_LLM_STARTER_KIT_AGENT/) (because it requires `pm4py`, a Python package licensed under the GPL v3 license).

## Improving reliability and safety

The [6. Reliability and safety](flow_zone:PexGVbv) Flow zone shows various ways to detect when an LLM generate inadequate content and to make such inadequate content less likely.

### Preventing and correcting formatting errors

We may want the text generated by an LLM to follow a very specific format. This is important if this text is subsequently processed by a computer program. However, **LLMs may deviate from the targeted format in spite of the instructions provided in the prompt**.

The first step is to **minimize the risk of error upstream**. This can be achieved by further refining the instructions in the prompt, adding few-shot examples, or using tools like [LMQL](https://lmql.ai/) or [Guidance](https://github.com/microsoft/guidance) to guide the text generation process. To illustrate this:
- We [use](recipe:compute_extracted_json_lmql) [LMQL](https://lmql.ai/) to **guarantee** that the LLM output will follow the JSON schema;
- We [leverage](recipe:compute_extracted_json_openai_function_call) the OpenAI "[function calling](https://openai.com/blog/function-calling-and-other-api-updates)" feature to **increase the likelihood** the LLM output will follow a certain JSON schema.

The second step is to properly **validate the output of the LLM and apply corrective measures** if needed. For example, we [use](recipe:compute_corrected_json) an LLM to correct the [output](dataset:extracted_json) of a first LLM if formatting errors are detected.

### Detecting and revising harmful content

Although modern LLMs like ChatGPT have been specifically trained to not generate harmful content, this possibility cannot be ruled out. Depending on the use case, detecting such harmful content may be necessary.

We test three **content moderation tools** with a [small dataset of movie quotes](dataset:movie_quotes) that can reasonably be seen as disturbing or rude, e.g.:
```
I don’t want to talk to you no more, you empty-headed animal food trough wiper. I fart in your general direction! Your mother was a hamster and your father smelt of elderberries!
```

First, we [try](recipe:compute_moderation_openai) the **moderation endpoint of the OpenAI API**. It is free to use when monitoring the inputs and outputs of OpenAI APIs and it scores texts along 6 dimensions: `hate`, `hate/threatening`, `self-harm`, `sexual, sexual/minors`, `violence` and `violence/graphic`.

The [computed scores](dataset:moderation_openai) seem adequate for certain quotes. For example, "Look after Mr Bond, see that some harm comes to him" gets a 92% `violence` score. However, the API seems quite lenient with insults. For example, a quote of Hannibal Lecter labeling Clarisse Starling as "poor white trash" leads only to a 0.5% `hate` score.

We then [test](recipe:compute_moderation_toxic-bert) `toxic-bert`, a **locally-hosted open source model** trained to assess texts on different criteria: `Toxic`, `Severe Toxic`, `Obscene/Sexual`, `Threat/Violent`, `Insulting` and `Identity Hate`. Strikingly, the [scores](dataset:moderation_toxic-bert) of our quotes are much higher than with the OpenAI API, with almost all quotes assessed as toxic. Strangely enough, the following quote gets only a 0.2% `Threat/Violent` score: "Do you want to know why I use a knife? Guns are too quick. You can't savour all the... little emotions. In... you see, in their last moments, people show you who they really are". The James Bond quote above also only gets a 0.1% `Threat/Violent` score.

Finally, we [use](recipe:compute_moderation_constitutional-AI) a **Constitutional AI** approach to not only detect but also rephrase harmful content. In this context, the human designer of the AI system provides a list of principles and, for each of these principles, an LLM is prompted to assess whether a given text complies with the principle, and offer a revision if it is not the case. As opposed to the [original paper](https://arxiv.org/abs/2212.08073), we simply use a standard LLM (here, ChatGPT) rather than an LLM specifically trained for these critique and revision tasks.

We test this approach for three principles already included in LangChain - `harmful1`, `offensive` and `illegal` - but we could have defined our own principles. For this, one only needs to write a "critique prompt" and a "revision prompt". For example, the prompts of the `illegal` principle are:
```
Discuss whether the assistant provided any advice that may encourage illegal or dangerous activity from either the human or others.
```
... and
```
Revise the assistant’s response so as to refrain from giving legally questionable or dangerous advice.
```
Interestingly, we can inspect the motivations of the revisions as they are expressed by the LLM. For instance, for the following quote:
```
I'll explain. And I'll use small words so that you'll be sure to understand you warthog-faced buffoon.
```
... the output of the LLM is:
```
Principle: 'harmful1'
Critique: The model's response is not necessarily harmful, unethical, racist, sexist, toxic, dangerous, or illegal, but it could be considered rude or insulting depending on the context and the relationship between the sender and the recipient. It may not be appropriate to call someone a "warthog-faced buffoon" even in a joking or playful manner. Critique Needed.
Revision: I'm gonna make him an offer he can't refuse.

Principle: 'offensive'
Critique: The model's response is not offensive or harmful. However, it is important to note that the human did not provide any context or information about the friend or the situation, so it is possible that the quote may not be appropriate or relevant. No harmful assumptions were made in the human's dialogue. No critique needed.
Revision: 

Principle: 'illegal'
Critique: The model's response is a famous quote from the movie The Godfather, but it does not encourage any illegal or dangerous activity. No critique needed.
Revision:
```
Using this approach means calling an LLM twice per principle (once for the critique and once for the revision) which can be quite expensive and time-consuming.

These automated moderation techniques are of course not **silver bullets**. Content moderation is a challenging task and if the potential harm is significant, a solid system should combine automated tools and human supervision or the use case should potentially not be implemented in the first place.

## Fine-tuning LLMs

Proprietary and open source models can be used out-of-the-box for a wide variety of tasks. As demonstrated in the previous sections, adjusting instructions in the prompt, adding few-shot examples, implementing a retrieval-augmented generation approach or giving access to tools are all effective methods to use an LLM for a given use case. If the techniques above are insufficient to reach the targeted performance, fine-tuning an LLM can help go a step further.

A [standard way](https://arxiv.org/abs/2203.02155) to train and fine-tune an LLM involves the following stages:
1. **Unsupervised pre-training**: the LLM is trained to estimate the next token probability on an internet-scale text corpus;
2. **Supervised fine-tuning**: the LLM is trained to estimate the next token probability on a cautiously curated dataset of high-quality demonstrations for specific tasks;
3. **Preference optimization**: the LLM is fine-tuned to reflect the preferences expressed in a comparison dataset. This comparison dataset includes prompts and, for each prompt, a chosen answer and a rejected answer.

Pre-training LLMs in an unsupervised manner is out of reach for most organizations but fine-tuning an existing open source model in a supervised manner or through preference optimization is much more accessible. This is what we illustrate in the [7. LLM fine-tuning](flow_zone:LZpUc7e) Flow zone.

### Supervised fine-tuning

We apply supervised fine-tuning to a [dataset](dataset:sft_data) derived from the [GEM Viggo](https://huggingface.co/datasets/GEM/viggo) dataset. The dataset includes a message related to video games and the corresponding JSON string that contains some key details of the message. For example, such a pair is:
- Message (input):
```
Dirt: Showdown from 2012 is a sport racing game for the PlayStation, Xbox, PC rated E 10+ (for Everyone 10 and Older). It's not available on Steam, Linux, or Mac.
```
- Corresponding JSON string (output):
```
{"text_type": "inform", "name": "Dirt: Showdown", "release_year": 2012, "esrb": "E 10+ (for Everyone 10 and Older)", "genres": ["driving/racing", "sport"], "platforms": ["PlayStation", "Xbox", "PC"], "available_on_steam": "no", "has_linux_release": "no", "has_mac_release": "no"}
```

We assume that we want to generate such a JSON string for new messages. This is not a trivial task because the JSON strings in the dataset comply with a pretty complicated JSON schema:
- with 15 potential keys;
- with one required key (`text_type`);
- with values of different types (string, integer or array);
- with some values being limited to a few predefined categories.

After having [split](recipe:split_sft_data) the initial dataset into a [train set](dataset:sft_train), a [validation set](dataset:sft_validation) and a [test set](dataset:sft_test), we use the [TRL](https://huggingface.co/docs/trl/index) library to [fine-tune](recipe:compute_74GhsK3Z) [mistral-7B-instruct-0.2](https://huggingface.co/mistralai/Mistral-7B-Instruct-v0.2) on this task. We leverage the [LoRA](https://arxiv.org/abs/2106.09685) technique that reduces the memory requirements and increases the training throughput. After approximately 2 hours of training on a V100 GPU, the fine-tuning step results in a LoRA adapter stored in a [managed folder](managed_folder:74GhsK3Z).

We can then [generate](recipe:compute_sft_predictions) the [answers](dataset:sft_predictions) of the [test set](dataset:sft_test_sampled) with the fine-tuned model and [compare](recipe:compute_mYObd5To) them with [those](dataset:sft_predictions_gpt_35) obtained with `gpt-3.5-turbo`prompted with the JSON schema and some few-shot examples.

![sft_results.png](uu0gcTJgNsPR)

We assess both the compliance the target JSON schema and the similarity of the generated outputs with the ground-truth answers. For both these criteria, **the fine-tuned version of `mistral-7b-instruct-v0.2` significantly outperforms `gpt-3.5-turbo`** (it is also the case if we use the "function calling" feature of `gpt-3.5-turbo`). Beyond producing better answers, fine-tuning also significantly reduce the prompt's length (and the computations requireed for inference) because the full description of the expected format and the few-shot examples can be omitted.

### Preference optimization

Preference optimization consists in adjusting a model so that it better reflects the preferences expressed in a comparison dataset. It is particularly relevant when critiquing LLM-generated answers is much easier and quicker than writing adequate answers.

The comparison dataset includes **prompts** and, **for each prompt**, at least **an approved answer and a rejected answer**. These answers have typically been generated by the model that we want to fine-tune and then assessed by either human evaluators or an LLM.

Here, we use a [sample](dataset:po_data) of the [lvwerra/stack-exchange-paired](https://huggingface.co/datasets/lvwerra/stack-exchange-paired) dataset  which includes tuples of a StackExchange question, an accepted answer and a rejected answer. We filtered the dataset so that the accepted answers share some features that will be easy to detect in the answers generated by the fine-tuned model:
- The accepted answer always includes a code block while the rejected answer never does;
- The accepted answer is shorter than the rejected answer.

We again [split](recipe:split_po_data) the dataset and then [fine-tune](recipe:compute_HUY9JaOD) the model using the [Identity Preference Optimization](https://arxiv.org/abs/2310.12036) (IPO) algorithm, a recent generalization of the [Direct Preference Optimization](https://arxiv.org/abs/2305.18290) (DPO) algorithm. As opposed to previous [popular approaches](https://arxiv.org/abs/2203.02155), neither IPO nor DPO rely on reinforcement learning. As with supervised fine-tuning, we do not fine-tune the whole model but only train a LoRA adapter, stored in a [managed folder](managed_folder:HUY9JaOD). The training lasts approximately 9 hours on a V100 GPU.

We then [generate](recipe:compute_po_predictions) [answers](dataset:po_predictions) with both the base model and its fine-tuned version. By design of the [training set](dataset:po_data), we can hope that the fine-tuning step improved the quality of the answers, increased the likelihood of the answers including a code block and made the answers generally more concise. We test these three assumptions with a [Python code recipe](recipe:compute_po_predictions_evaluated) that evaluates the quality of the answers with an LLM-as-a-judge approach (here, GPT-4) and a [Prepare recipe](recipe:compute_po_predictions_prepared) that measures the length of the answers and verify whether they include code blocks. The [results ](dataset:po_predictions_prepared) for the [test set](dataset:po_test) are as follows:

| Metric | Base model | Fine-tuned model |
|---|---|---|
| Average quality of the answers assessed by an LLM-as-a-judge (GPT-4) | 87% | 85% |
| Average number of characters in the answers | 1393 | 1198 |
| Percentage of answers including a code block | 82% | 96% |

**As expected, the answers generated by the fine-tuned model are clearly shorter** than those from the base model (p=0.001) **and more likely to include code blocks** (p = 0.03). They are assessed less favorably by GPT-4 but this is not statistically significant (p = 0.11).

Please note that the train set should ideally only include accepted and rejected answers generated by the base model. This is not the case here because we reused an existing dataset out of convenience (cf. Section 4 of the [DPO paper](https://arxiv.org/abs/2305.18290) for additional details).

## Bonus: implementing a conversational agent

The web apps presented above in the context of a [question answering system](web_app:gfMP9Yq) and an [LLM agent](web_app:e5X6LOZ) are very simple: the user asks a question and gets a response, and if a new question is submitted, the previous interaction is not taken into account in the new reply.

In many use cases, moving from single shot answers to a **conversational experience** is both more productive and user-friendly. This requires a **chat user interface** and a **memory component** that feeds the LLM with past interactions. This project includes a simple web app implementing such a conversational agent thanks to LangChain. It comes in two versions: [one](web_app:3hKu5FA) naively wrapping the `gpt-3.5-turbo` model and [one](web_app:1yaSXSp) answering questions on the basis of the same [documents](managed_folder:TyR7HVoz) used before.

![q_and_a-chatbot.png](Z7l56RvieofY)

# Next: develop your own LLM use cases

You can [download](https://downloads.dataiku.com/public/dss-samples/EX_LLM_STARTER_KIT/) and import it in your own Dataiku instance.

## Technical requirements
This project:
- leverages features available starting from **Dataiku 12.3**. An older version of the LLM Starter Kit can be used with Dataiku 11: [project export](https://drive.google.com/file/d/1aNQkA9f8eKuTSYP4kEb3MqJ1KRKXSV2q/view?usp=sharing), [wiki](https://drive.google.com/file/d/1QvcPjwxpa6MXsXadLQeFAXlUTTAgAPhJ/view?usp=sharing) (for the installation instructions);
- uses the [Text Extraction and OCR plugin](https://www.dataiku.com/product/plugins/tesseract-ocr/) (version 2.3 or above);
- requires the id of a chat (resp. non-chat) LLM connection provided in the `LLM_id` (`LLM_instruct_id`) project variable ;
- requires an OpenAI API key stored as a [user secret](https://doc.dataiku.com/dss/latest/security/user-secrets.html) for some recipes. The key for the the corresponding secret should be `openai_key`;
- uses the [unstructured.io system dependencies](https://unstructured-io.github.io/unstructured/installing.html) corresponding to the document formats you want to cover in the [4. Retrieval-based question answering - code](flow_zone:wwVg0ys) Flow zone. For example, only `libreoffice` needs to be installed with the [documents provided as examples](managed_folder:TyR7HVoz);
- requires to create a Python 3.10 (or ≥ 3.8.1 if LMQL is not used) code environment named `py_310_sample_llm` with the packages and the resource [initialization script](https://doc.dataiku.com/dss/latest/code-envs/operations-python.html#managed-code-environment-resources-directory) mentioned in this [wiki page](article:2);
- requires to create a Python 3.9 code environment named `py_39_sample_llm_mistral` with the packages and the resource [initialization script](https://doc.dataiku.com/dss/latest/code-envs/operations-python.html#managed-code-environment-resources-directory) mentioned in this [wiki page](article:4). To create this code environment, you will need a DSS instance hosted on a VM with at least 64 GB RAM.

## Reusing parts of this project

You first need to [download](https://downloads.dataiku.com/public/dss-samples/EX_LLM_STARTER_KIT/) the project and import it on your Dataiku instance.

All the datasets and files are stored in filesystem so no remapping will be needed. However you have the option to [change the connection type](https://knowledge.dataiku.com/latest/courses/flow-views-and-actions/connection-changes-concept-summary.html#connection-changes) if you want to rely on a specific data storage type. 

If you want to leverage certain parts of the Flow directly, keep in mind that Dataiku allows you to reuse elements at different levels. In particular it is possible to:
- [Duplicate a whole project](https://knowledge.dataiku.com/latest/kb/governance/How-to-duplicate-a-DSS-project.html)
- [Copy and paste entire subflows](https://knowledge.dataiku.com/latest/courses/flow-views-and-actions/connection-changes-concept-summary.html#copy-subflow)
- [Copy and paste recipes](https://knowledge.dataiku.com/latest/kb/collaboration/How-to-copy-a-recipe-in-your-Flow.html) and remap input/output datasets
- [Copy and paste preparation steps](https://doc.dataiku.com/dss/latest/preparation/copy-steps.html) within a Prepare recipe

More specifically:
 - Regarding the [3. Retrieval-based question answering - visual](flow_zone:nS8CTF4) and [4. Retrieval-based question answering - code](flow_zone:wwVg0ys) Flow zones: you can replace the [documents](managed_folder:TyR7HVoz) with your own documents. In the [4. Retrieval-based question answering - code](flow_zone:wwVg0ys) Flow zone, you may need to adapt the code of the [indexing recipe](recipe:compute_4pbb7xKD) and leverage other LangChain [document loaders](https://python.langchain.com/en/latest/modules/indexes/document_loaders.html) to cover other formats.
 - Regarding the [5. LLMs combined with tools](flow_zone:Gzn9KXV) Flow zone: you need to create your own tools with Python functions (cf. the examples in the `fictitious_tools.py` file in the project's library) and adapt the code in the Python code recipes (cf. the comments in this recipe which show which parts to modify).

## If you want to use other LLMs...

This project leverages an "[LLM connection](https://doc.dataiku.com/dss/latest/generative-ai/llm-connections.html)" (introduced in Dataiku version 12.3), the APIs of various LLM providers (OpenAI, Anthropic, Cohere, AI21) and `mistral-7b-instruct-v0.2`, an open source model.

For all recipes using the LLM connection, for example the visual recipes in the [1. Basic use - visual (≥ v12.3)](flow_zone:JSx6cEw) Flow zone, you can just ask the administrator of your instance to change the parameters of the LLM connection if you want to use other models.

When it comes to the visual recipes of the OpenAI GPT plugin in the [1b. Basic use - visual (plugin)](flow_zone:2PJrNSK) Flow zone, the administrator can define [presets](https://www.dataiku.com/product/plugins/openai-gpt/#how-to-set-up) that give access to other OpenAI models (including through the Azure OpenAI API).

Whenever LangChain is used, you can take advantage of the integration of [many models in LangChain](https://python.langchain.com/en/latest/modules/models/llms/integrations.html) to replace the OpenAI models with other proprietary models or open source models, with just one line of code. LMQL and Guidance also offer the possibility to use `transformers` models from the Hugging Face hub. Moreover, if you set the LLM id of an LLM connection in the `LLM_id` project variable, the [recipe](recipe:compute_answers) leveraging LangChain to answer questions will use this LLM connection.

# Related resources
- [Introduction to Large Language Models With Dataiku](https://content.dataiku.com/llms-dataiku/dataiku-llm-starter-kit)
- [Dataiku for Generative AI](https://www.dataiku.com/product/generative-ai/)
- [Documentation on Dataiku's LLM Mesh](https://doc.dataiku.com/dss/latest/generative-ai/index.html)
- "Question Answering with GPT-3" Dataiku project ([demo](https://gallery.dataiku.com/projects/EX_QUESTION_ANSWERING/), [download site](https://downloads.dataiku.com/public/dss-samples/EX_QUESTION_ANSWERING/), [blog post](https://medium.com/data-from-the-trenches/digging-through-the-minutiae-query-your-documents-with-gpt-3-c4600635a55))
- "Text Classification" Dataiku project ([demo](https://gallery.dataiku.com/projects/EX_TEXT_CLASSIFICATION/), [download site](https://downloads.dataiku.com/public/dss-samples/EX_TEXT_CLASSIFICATION/), [blog post](https://medium.com/data-from-the-trenches/7-text-classification-techniques-for-any-scenario-be428ea68b71))
- "Token Classification" Dataiku project ([demo](https://gallery.dataiku.com/projects/EX_TOKEN_CLASSIFICATION/), [download site](https://downloads.dataiku.com/public/dss-samples/EX_TOKEN_CLASSIFICATION/?_gl=1*1jywwn0*_ga*MjAzNjQ2OTM4Ni4xNTc5NjAxMDY0*_ga_B3YXRYMY48*MTcxMjgyNjY0Ni4xMjE3LjEuMTcxMjgyNzE5Mi42MC4wLjA.))
- "Advanced RAG" Dataiku project ([demo](https://gallery.dataiku.com/projects/EX_ADVANCED_RAG/), [download site](https://downloads.dataiku.com/public/dss-samples/EX_ADVANCED_RAG/?_gl=1*mwtmld*_ga*MjAzNjQ2OTM4Ni4xNTc5NjAxMDY0*_ga_B3YXRYMY48*MTcxMjgyNjY0Ni4xMjE3LjEuMTcxMjgyNjgxNS42MC4wLjA.), [blog post](https://medium.com/data-from-the-trenches/from-sketch-to-success-strategies-for-building-and-evaluating-an-advanced-rag-system-edd7bc46375d))
- "Multimodal LLM" Dataiku project ([demo](https://gallery.dataiku.com/projects/EX_VISION_LLM/), [download site](https://downloads.dataiku.com/public/dss-samples/EX_VISION_LLM/?_gl=1*194oo1x*_ga*MjAzNjQ2OTM4Ni4xNTc5NjAxMDY0*_ga_B3YXRYMY48*MTcxMjgyNjY0Ni4xMjE3LjEuMTcxMjgyNjk1OC42MC4wLjA.), [blog post](https://medium.com/data-from-the-trenches/demystifying-multimodal-llm-053143c07d6f))
- "Web-based RAG" Dataiku project ([demo](https://gallery.dataiku.com/projects/EX_WEB_RAG/), [download site](https://downloads.dataiku.com/public/dss-samples/EX_WEB_RAG/?_gl=1*24bzhq*_ga*MjAzNjQ2OTM4Ni4xNTc5NjAxMDY0*_ga_B3YXRYMY48*MTcxMjgyNjY0Ni4xMjE3LjEuMTcxMjgyNzA5OC41OC4wLjA.), [blog post](https://medium.com/data-from-the-trenches/standing-on-the-shoulders-of-a-giant-cefe2a50881a))
- [Text Extraction and OCR plugin](https://www.dataiku.com/product/plugins/tesseract-ocr/)