# Specifications of the py_310_sample_llm code environment

## Packages

Below are the packages required in the `py_310_sample_llm` code environment. **Please note that you should also add `pm4py==2.7.11.3` if you want to use the evaluation recipe of the [companion project](https://gallery.dataiku.com/projects/EX_LLM_STARTER_KIT_AGENT/)** and that `pm4py` is licensed under the GPL v3 license.

```
openai==1.14.3
tiktoken==0.5.2
lmql==0.7.3
guidance==0.1.13
langchain==0.1.13
langchainhub==0.1.15
langchain-openai==0.1.2
wikipedia==1.4.0
torch==2.0.1
transformers==4.31.0
sentence-transformers==2.2.2
accelerate==0.21.0
evaluate==0.4.1
bert-score==0.3.13
faiss-cpu==1.7.4
pypdf==3.15.0
unstructured[local-inference]==0.9.1
tabulate==0.9.0
dash==2.11.1
dash-bootstrap-components==1.4.2
mlflow==2.11.3
textstat==0.7.3
spacy==3.7.0
https://github.com/explosion/spacy-models/releases/download/en_core_web_sm-3.7.0/en_core_web_sm-3.7.0.tar.gz
```

## Resource initialization script

These resources may not be needed if the corresponding recipes are not used (cf. comments in the script).

```
######################## Base imports #################################
import logging
import os
import shutil

from dataiku.code_env_resources import clear_all_env_vars
from dataiku.code_env_resources import grant_permissions
from dataiku.code_env_resources import set_env_path
from dataiku.code_env_resources import set_env_var
from dataiku.code_env_resources import update_models_meta

# Set-up logging
logging.basicConfig()
logger = logging.getLogger("code_env_resources")
logger.setLevel(logging.INFO)

# Clear all environment variables defined by a previously run script
clear_all_env_vars()

######################## NLTK ##################################################
# Useful only for the 3. "Retrieval-based question answering" Flow zone
################################################################################

# Set NLTK data directory
set_env_path("NLTK_DATA", "nltk_data")

# Import NLTK
import nltk

# Download model: automatically managed by NLTK, does not download
# anything if model is already in NLTK_DATA.
nltk.download('punkt', download_dir=os.environ["NLTK_DATA"])
nltk.download('averaged_perceptron_tagger', download_dir=os.environ["NLTK_DATA"])

######################## Hugging Face #################################

# Set HuggingFace cache directory
set_env_path("HF_HOME", "huggingface")
hf_home_dir = os.getenv("HF_HOME")

################################################################################
# Useful only for the 5. Safety and reliability Flow zone
################################################################################
from transformers import AutoTokenizer, AutoModelForSequenceClassification
tokenizer = AutoTokenizer.from_pretrained('unitary/toxic-bert')
model = AutoModelForSequenceClassification.from_pretrained('unitary/toxic-bert')

# Grant everyone read access to pretrained models in the HF_HOME folder
# (by default, only readable by the owner)
grant_permissions(hf_home_dir)

################################################################################
# Useful only for the 3. "Retrieval-based question answering" Flow zone
################################################################################

import evaluate
bertscore = evaluate.load("bertscore")
results = bertscore.compute(
    predictions=["a"],
    references=["b"],
    model_type="distilbert-base-uncased"
)

######################## Sentence Transformers #################################
# Useful only for the 3. "Retrieval-based question answering" Flow zone
################################################################################

# Set sentence_transformers cache directory
set_env_path("SENTENCE_TRANSFORMERS_HOME", "sentence_transformers")

import sentence_transformers

# Download pretrained models
MODELS_REPO_AND_REVISION = [
    ("sentence-transformers/msmarco-distilbert-cos-v5", "97bf29337ec20da8a1fb1ff2bd5555de5b566baf"),
]

sentence_transformers_cache_dir = os.getenv("SENTENCE_TRANSFORMERS_HOME")
for (model_repo, revision) in MODELS_REPO_AND_REVISION:
    logger.info("Loading pretrained SentenceTransformer model: {}".format(model_repo))
    model_path = os.path.join(sentence_transformers_cache_dir, model_repo.replace("/", "_"))

    # This also skips same models with a different revision
    if not os.path.exists(model_path):
        model_path_tmp = sentence_transformers.util.snapshot_download(
            repo_id=model_repo,
            revision=revision,
            cache_dir=sentence_transformers_cache_dir,
            library_name="sentence-transformers",
            library_version=sentence_transformers.__version__,
            ignore_files=["flax_model.msgpack", "rust_model.ot", "tf_model.h5",],
        )
        os.rename(model_path_tmp, model_path)
    else:
        logger.info("Model already downloaded, skipping")

# Add sentence embedding models to the code-envs models meta-data
# (ensure that they are properly displayed in the feature handling)
update_models_meta()
# Grant everyone read access to pretrained models in sentence_transformers/ folder
# (by default, sentence transformers makes them only readable by the owner)
grant_permissions(sentence_transformers_cache_dir)
```